import IPython.display as display
import numpy as np
import PIL.Image
import tensorflow as tf


def download(url, max_dim=None):
    """Download an image and load it as a NumPy array."""
    name = url.split("/")[-1]
    image_path = tf.keras.utils.get_file(name, origin=url)
    img = PIL.Image.open(image_path)
    if max_dim:
        img.thumbnail((max_dim, max_dim))
    return np.array(img)


def deprocess(img):
    """Normalize image for display."""
    img = 255 * (img + 1.0) / 2.0
    return tf.cast(img, tf.uint8)


def show(img):
    """Display an image."""
    display.display(PIL.Image.fromarray(np.array(img)))


class DeepDream(tf.Module):
    def __init__(self, model):
        self.model = model

    @tf.function(
        input_signature=(
            tf.TensorSpec(shape=[None, None, 3], dtype=tf.float32),
            tf.TensorSpec(shape=[], dtype=tf.int32),
            tf.TensorSpec(shape=[], dtype=tf.float32),
        )
    )
    def __call__(self, img, steps, step_size):
        print("Tracing DeepDream computation graph...")
        loss = tf.constant(0.0)
        for n in tf.range(steps):
            with tf.GradientTape() as tape:
                tape.watch(img)
                loss = calc_loss(img, self.model)
                gradients = tape.gradient(loss, img)
            gradients /= tf.math.reduce_std(gradients) + 1e-8
            img = img + gradients * step_size
            img = tf.clip_by_value(img, -1, 1)

        return loss, img


def calc_loss(img, model):
    """Calculate the DeepDream loss by maximizing activations."""
    img_batch = tf.expand_dims(img, axis=0)
    layer_activations = model(img_batch)
    if len(layer_activations) == 1:
        layer_activations = [layer_activations]

    return tf.reduce_sum([tf.math.reduce_mean(act) for act in layer_activations])


def run_deep_dream_simple(img, dream_model, steps=100, step_size=0.01):
    img = tf.keras.applications.inception_v3.preprocess_input(img)
    img = tf.convert_to_tensor(img)
    base_shape = tf.shape(img)[:-1]

    step_size = tf.convert_to_tensor(step_size)
    steps_remaining = steps
    step = 0
    while steps_remaining:
        if steps_remaining > 100:
            run_steps = tf.constant(100)
        else:
            run_steps = tf.constant(steps_remaining)
        steps_remaining -= run_steps
        step += run_steps

        loss, img = dream_model(img, run_steps, tf.constant(step_size))

        # display.clear_output(wait=True)
        show(deprocess(img))
        print("Step {}, loss {}".format(step, loss))

    img = tf.image.resize(img, base_shape)
    result = deprocess(img)
    # display.clear_output(wait=True)
    show(result)

    return result


def run_deep_dream_octaved(img, dream_model, steps_per_octave=100, step_size=0.01):
    OCTAVE_SCALE = 1.30

    img = tf.constant(np.array(img))
    base_shape = tf.shape(img)[:-1]
    float_base_shape = tf.cast(base_shape, tf.float32)

    for n in range(-2, 3):
        new_shape = tf.cast(float_base_shape * (OCTAVE_SCALE**n), tf.int32)

        img = tf.image.resize(img, new_shape).numpy()
        img = run_deep_dream_simple(
            img=img,
            dream_model=dream_model,
            steps=steps_per_octave,
            step_size=step_size,
        )

    # display.clear_output(wait=True)
    img = tf.image.resize(img, base_shape)
    img = tf.image.convert_image_dtype(img / 255.0, dtype=tf.uint8)
    show(img)

    return img


def random_roll(img, maxroll):
    # Randomly shift the image to avoid tiled boundaries.
    shift = tf.random.uniform(
        shape=[2], minval=-maxroll, maxval=maxroll, dtype=tf.int32
    )
    img_rolled = tf.roll(img, shift=shift, axis=[0, 1])
    return shift, img_rolled


class TiledGradients(tf.Module):
    def __init__(self, model):
        self.model = model

    @tf.function(
        input_signature=(
            tf.TensorSpec(shape=[None, None, 3], dtype=tf.float32),
            tf.TensorSpec(shape=[2], dtype=tf.int32),
            tf.TensorSpec(shape=[], dtype=tf.int32),
        )
    )
    def __call__(self, img, img_size, tile_size=512):
        shift, img_rolled = random_roll(img, tile_size)

        # Initialize the image gradients to zero.
        gradients = tf.zeros_like(img_rolled)

        # Skip the last tile, unless there's only one tile.
        xs = tf.range(0, img_size[1], tile_size)[:-1]
        if not tf.cast(len(xs), bool):
            xs = tf.constant([0])
        ys = tf.range(0, img_size[0], tile_size)[:-1]
        if not tf.cast(len(ys), bool):
            ys = tf.constant([0])

        for x in xs:
            for y in ys:
                # Calculate the gradients for this tile.
                with tf.GradientTape() as tape:
                    # This needs gradients relative to `img_rolled`.
                    # `GradientTape` only watches `tf.Variable`s by default.
                    tape.watch(img_rolled)

                    # Extract a tile out of the image.
                    img_tile = img_rolled[y : y + tile_size, x : x + tile_size]
                    loss = calc_loss(img_tile, self.model)

                # Update the image gradients for this tile.
                gradients = gradients + tape.gradient(loss, img_rolled)

        # Undo the random shift applied to the image and its gradients.
        gradients = tf.roll(gradients, shift=-shift, axis=[0, 1])

        # Normalize the gradients.
        gradients /= tf.math.reduce_std(gradients) + 1e-8

        return gradients


def run_deep_dream_rolled(
    img,
    get_tiled_gradients,
    steps_per_octave=100,
    step_size=0.01,
    octaves=range(-2, 3),
    octave_scale=1.3,
):
    base_shape = tf.shape(img)
    img = tf.keras.utils.img_to_array(img)
    img = tf.keras.applications.inception_v3.preprocess_input(img)

    initial_shape = img.shape[:-1]
    img = tf.image.resize(img, initial_shape)
    for octave in octaves:
        # Scale the image based on the octave
        new_size = tf.cast(tf.convert_to_tensor(base_shape[:-1]), tf.float32) * (
            octave_scale**octave
        )
        new_size = tf.cast(new_size, tf.int32)
        img = tf.image.resize(img, new_size)

        for step in range(steps_per_octave):
            gradients = get_tiled_gradients(img, new_size)
            img = img + gradients * step_size
            img = tf.clip_by_value(img, -1, 1)

            if step % 10 == 0:
                # display.clear_output(wait=True)
                show(deprocess(img))
                print("Octave {}, Step {}".format(octave, step))

    result = deprocess(img)
    img = tf.image.resize(img, base_shape)
    img = tf.image.convert_image_dtype(img / 255.0, dtype=tf.uint8)
    show(img)
    return result


def main():
    url = (
        "https://storage.googleapis.com/download.tensorflow.org/"
        "example_images/YellowLabradorLooking_new.jpg"
    )

    original_img = download(url, max_dim=500)
    show(original_img)

    display.display(
        display.HTML(
            'Image cc-by: <a href="https://commons.wikimedia.org/wiki/'
            'File:Felis_catus-cat_on_snow.jpg">Von.grzanka</a>'
        )
    )

    base_model = tf.keras.applications.InceptionV3(
        include_top=False, weights="imagenet"
    )

    names = ["mixed3", "mixed5"]
    layers = [base_model.get_layer(name).output for name in names]

    dream_model = tf.keras.Model(inputs=base_model.input, outputs=layers)

    deepdream = DeepDream(dream_model)


    # Single Octave
    run_deep_dream_simple(
        img=original_img, dream_model=deepdream, steps=100, step_size=0.01
    )

    # Multi-Octave
    run_deep_dream_octaved(
        img=original_img, dream_model=deepdream, steps_per_octave=50, step_size=0.01
    )
    
    # Multi-Octave with Tiling
    get_tiled_gradients = TiledGradients(dream_model)
    original_img = tf.convert_to_tensor(original_img)
    run_deep_dream_rolled(
        img=original_img, get_tiled_gradients=get_tiled_gradients, steps_per_octave=50, step_size=0.01
    )


if __name__ == "__main__":
    main()



import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf
from moviepy.video.fx import AccelDecel
from moviepy.video.VideoClip import DataVideoClip
from tensorflow import keras
from tqdm import trange

from dreamify.utils.configure import Config

config: Config = None


def configure_settings(
    feature_extractor,
    layer_settings,
    original_shape,
    enable_framing,
    frames_for_vid,
    iterations,
):
    global config
    config = Config(
        feature_extractor=feature_extractor,
        layer_settings=layer_settings,
        original_shape=original_shape,
        enable_framing=enable_framing,
        frames_for_vid=frames_for_vid,
        max_frames_to_sample=iterations,
    )


def preprocess_image(image_path):
    img = keras.utils.load_img(image_path)
    img = keras.utils.img_to_array(img)
    img = np.expand_dims(img, axis=0)
    img = keras.applications.inception_v3.preprocess_input(img)
    return img


def deprocess_image(img):
    img = img.reshape((img.shape[1], img.shape[2], 3))
    img /= 2.0
    img += 0.5
    img *= 255.0
    img = np.clip(img, 0, 255).astype("uint8")
    return img


def compute_loss(input_image):
    features = config.feature_extractor(input_image)
    loss = tf.zeros(shape=())
    for name in features.keys():
        coeff = config.layer_settings[name]
        activation = features[name]
        loss += coeff * tf.reduce_mean(tf.square(activation[:, 2:-2, 2:-2, :]))
    return loss


@tf.function
def gradient_ascent_step(image, learning_rate):
    with tf.GradientTape() as tape:
        tape.watch(image)
        loss = compute_loss(image)
    grads = tape.gradient(loss, image)
    grads = tf.math.l2_normalize(grads) + 1e-8
    image += learning_rate * grads
    image = tf.clip_by_value(image, -1, 1)
    return loss, image


def gradient_ascent_loop(image, iterations, learning_rate, max_loss=None):
    global config

    for i in trange(
        iterations, desc="Gradient Ascent", unit="step", ncols=75, mininterval=0.1
    ):
        loss, image = gradient_ascent_step(image, learning_rate)

        if max_loss is not None and loss > max_loss:
            print(
                f"\nTerminating early: Loss ({loss:.2f}) exceeded max_loss ({max_loss:.2f}).\n"
            )
            break

        if (
            config.enable_framing
            and config.curr_frame_idx < config.max_frames_to_sample - 1
        ):
            frame = tf.image.resize(image, config.original_shape)
            frame = deprocess_image(image.numpy())
            config.frames_for_vid.append(frame)
            config.curr_frame_idx += 1

    return image


def to_video(output_path, duration, fps=60):
    global config

    upsample()

    print(f"Number of images to frame: {len(config.frames_for_vid)}")

    vid = DataVideoClip(config.frames_for_vid, lambda x: x, fps=fps)
    vid = AccelDecel(new_duration=duration).apply(vid)
    vid.write_videofile(output_path)

    config = Config()  # Reset the configuration


@tf.function
def interpolate_frames(frame1, frame2, num_frames):
    alphas = tf.linspace(0.0, 1.0, num_frames + 2)[1:-1]  # Avoid 0 and 1
    return tf.cast(
        (1 - alphas[:, None, None, None]) * frame1
        + alphas[:, None, None, None] * frame2,
        tf.uint8,
    )


def upsample():
    global config
    NUM_FRAMES_TO_INSERT = 60

    new_frames = []

    # Upsample via frame-frame interpoliation
    for i in range(len(config.frames_for_vid) - 1):
        frame1 = tf.convert_to_tensor(config.frames_for_vid[i], dtype=tf.float32)
        frame2 = tf.convert_to_tensor(config.frames_for_vid[i + 1], dtype=tf.float32)

        new_frames.append(config.frames_for_vid[i])
        new_frames.extend(
            interpolate_frames(frame1, frame2, NUM_FRAMES_TO_INSERT).numpy()
        )

    new_frames.extend(
        [config.frames_for_vid[-1]] * 60 * 3
    )  # Lengthen end frame by 3 secs
    config.frames_for_vid = new_frames


def show(img):
    plt.imshow(np.squeeze(img))
    plt.axis("off")
    plt.show()


__all__ = [
    configure_settings,
    preprocess_image,
    deprocess_image,
    gradient_ascent_loop,
    to_video,
    show,
]
